@@ -3,7 +3,7 @@
 ;; Copyright (C) 2015-2017  Jorgen Schaefer <contact@jorgenschaefer.de>
 ;; Copyright (C) 2018-2024  Ola Nilsson <ola.nilsson@gmail.com>
 
-;; Version: 1.35
+;; Version: 1.36
 ;; Author: Jorgen Schaefer <contact@jorgenschaefer.de>
 ;; Maintainer: Ola Nilsson <ola.nilsson@gmail.com>
 ;; Package-Requires: ((emacs "24.4"))
@@ -69,10 +69,12 @@
               'buttercup-internals-error)
 
 (eval-and-compile
-  (when (fboundp 'oclosure-define)      ;Emacs≥29
+  (if (fboundp 'oclosure-define)      ;Emacs≥29
     (oclosure-define (buttercup--thunk (:predicate buttercup--thunk-p))
       "An elisp expression as a function and original code."
-      expr)))
+      expr)
+    (defalias 'buttercup--thunk-p #'ignore
+      "Always return nil when Oclosures are not available.")))
 
 (defun buttercup--enclosed-expr (fun)
   "Given a FUN `buttercup-thunk', return its unevaluated expression.
@@ -81,15 +83,13 @@ For Emacs < 29:
 The function MUST be byte-compiled or have one of the following
 forms:
 
-\(closure (ENVLIST) () (quote EXPR) (buttercup--mark-stackframe) EXPANDED)
-\(lambda () (quote EXPR) (buttercup--mark-stackframe) EXPR)
+\(closure (ENVLIST) () (quote EXPR) EXPANDED)
+\(lambda () (quote EXPR) EXPR)
 
 and the return value will be EXPR, unevaluated. The quoted EXPR
 is useful if EXPR is a macro call, in which case the `quote'
 ensures access to the un-expanded form."
-  (cl-assert (functionp fun) t "Expected FUN to be a function")
-  (if (and (fboundp 'buttercup--thunk-p)      ;Emacs≥29
-           (buttercup--thunk-p fun))
+  (if (buttercup--thunk-p fun)
       (buttercup--thunk--expr fun)
   (pcase fun
     ;; This should be the normal case, a closure with unknown enclosed
@@ -98,7 +98,7 @@ ensures access to the un-expanded form."
     ;; * the stackframe marker
     ;; * the macroexpanded original expression
     (`(closure ,(pred listp) nil
-        (quote ,expr) (buttercup--mark-stackframe) ,_expanded)
+        (quote ,expr) ,_expanded)
      expr)
     ;; This a when FUN has not been evaluated.
     ;; Why does that happen?
@@ -107,7 +107,7 @@ ensures access to the un-expanded form."
     ;; * the stackframe marker
     ;; * the expanded expression
     (`(lambda nil
-        (quote ,expr) (buttercup--mark-stackframe) ,_expanded)
+        (quote ,expr) ,_expanded)
      expr)
     ;; This is when FUN has been byte compiled, as when the entire
     ;; test file has been byte compiled. Check that it has an empty
@@ -185,30 +185,31 @@ Does not have the IGNORE-MISSING and SPLIT parameters."
 (define-error 'buttercup-pending "Buttercup test is pending" 'buttercup-error-base)
 
 (defun buttercup--wrap-expr (expr)
-  "Wrap EXPR in a `buttercup--thunk' to be used by `buttercup-expect'."
+  "Wrap EXPR in a `buttercup--thunk' to be used by `buttercup-expect'.
+This function is only usable from within the buttercup `expect' macro."
   (if (fboundp 'oclosure-lambda)        ;Emacs≥29
       `(oclosure-lambda (buttercup--thunk (expr ',expr)) ()
-         (buttercup--mark-stackframe)
          ,expr)
     `(lambda ()
        (quote ,expr)
-       (buttercup--mark-stackframe)
        ,expr)))
 
 (defmacro expect (arg &optional matcher &rest args)
   "Expect a condition to be true.
 
-This macro knows three forms:
+This macro knows two forms:
 
-\(expect ARG :MATCHER ARGS...)
-  Fail the current test if the matcher does not match these arguments.
+\(expect ARG :MATCHER [ARGS...])
+  Fail the current test if the MATCHER does not match these arguments.
+  The correct number of arguments depend on the MATCHER.
   See `buttercup-define-matcher' for more information on matchers.
 
-\(expect (function ARG...))
-  Fail the current test if the function call does not return a true value.
-
 \(expect ARG)
-  Fail the current test if ARG is not true."
+  Fail the current test if ARG is not true.
+  This is the same as (expect ARG :to-be-truthy).
+
+All arguments can be a value or a single Lisp form. The evaluation of
+the arguments is delayed until the containing spec is executed."
   `(buttercup-expect ,(buttercup--wrap-expr arg)
                      ,(or matcher :to-be-truthy)
                      ,@(mapcar #'buttercup--wrap-expr args)))
@@ -219,24 +220,14 @@ This macro knows three forms:
 See the macro documentation for details and the definition of
 ARG, MATCHER and ARGS."
   (cl-assert (cl-every #'buttercup--wrapper-fun-p (cons arg args)) t)
-  (if (not matcher)
-      (progn
-        (cl-assert (not args) t)
-        (when (not (funcall arg))
-          (buttercup-fail "Expected %S to be non-nil"
-                          (buttercup--enclosed-expr arg))))
-    (let ((result (buttercup--apply-matcher matcher (cons arg args))))
-      (if (consp result)
-          (when (not (car result))
-            (buttercup-fail "%s" (cdr result)))
-        (when (not result)
-          (buttercup-fail "Expected %S %S %s"
-                          (buttercup--enclosed-expr arg)
-                          matcher
-                          (mapconcat (lambda (obj)
-                                       (format "%S" (funcall obj)))
-                                     args
-                                     " ")))))))
+  (pcase (buttercup--apply-matcher (or matcher :to-be-truthy) (cons arg args))
+    (`(,result . ,message) (unless result (buttercup-fail message)))
+    (result (unless result (buttercup-fail "Expected %S %S %s"
+                                           (buttercup--enclosed-expr arg)
+                                           matcher
+                                           (mapconcat (lambda (obj)
+                                                        (format "%S" (funcall obj)))
+                                                      args " "))))))
 
 (defun buttercup-fail (format &rest args)
   "Fail the current test with the given description.
@@ -265,16 +256,16 @@ MESSAGE is omitted or nil show the condition form instead."
 
 (defmacro buttercup-define-matcher (matcher args &rest body)
   "Define a matcher named MATCHER to be used in `expect'.
-
+MATCHER is a keyword, for instance `:to-be'.
 ARGS is a list of the elements to match together.
-
 The BODY will receive ARGS as functions that can be called (using
 `funcall') to get their values. BODY should return either a
 simple boolean, or a cons cell of the form (RESULT . MESSAGE). If
 RESULT is nil, MESSAGE should describe why the matcher failed. If
 RESULT is non-nil, MESSAGE should describe why a negated matcher
 failed.
-BODY may start with a docstring."
+
+\(fn MATCHER ARGS [DOCSTRING] BODY...)"
   (declare (indent defun))
   `(put ,matcher 'buttercup-matcher
         (lambda ,args
@@ -616,6 +607,8 @@ See also `buttercup-define-matcher'."
   :expect-mismatch-phrase "Expected `%A' < %b, but `%A' was %a.")
 
 (buttercup-define-matcher :to-be-close-to (a b precision)
+  "Verify that |A-B| < 10^-PRECISION.
+\(expect A :to-be-close-to B PRECISION)"
   (cl-destructuring-bind
       (precision (a-expr . a) (_b-expr . b))
       (cons (funcall precision)
@@ -1015,7 +1008,6 @@ most probably including one or more calls to `expect'."
       `(buttercup-it ,description
          (lambda ()
            (buttercup-with-converted-ert-signals
-             (buttercup--mark-stackframe)
              ,@body)))
     `(buttercup-xit ,description)))
 
@@ -2107,44 +2099,87 @@ ARGS according to `debugger'."
                ;; args is (error (signal . data) ....) where the tail
                ;; may be empty
                (cl-destructuring-bind (signal-type . data) (cl-second args)
-                 (unless (eq signal-type 'buttercup-pending)
-                   (buttercup--backtrace))))))
-
-(defalias 'buttercup--mark-stackframe #'ignore
-  "Marker to find where the backtrace start.")
+                 (cl-case signal-type
+                   ((buttercup-pending buttercup-failed))
+                   (otherwise (buttercup--backtrace)))))))
 
 (defun buttercup--backtrace ()
   "Create a backtrace, a list of frames returned from `backtrace-frame'."
-  ;; Read the backtrace frames from 0 (the closest) upward.
-  (cl-do* ((n 0 (1+ n))
-           (frame (backtrace-frame n) (backtrace-frame n))
-           (frame-list nil)
-           (in-program-stack nil))
+  ;; Read the backtrace frames from `buttercup--debugger' + 1  upward.
+  (cl-do* ((n 1 (1+ n))
+           (frame (backtrace-frame n #'buttercup--debugger)
+                  (backtrace-frame n #'buttercup--debugger))
+           (frame-list nil))
       ((not frame) frame-list)
-      ;; discard frames until (and including) `buttercup--debugger', they
-      ;; only contain buttercup code
-      (when in-program-stack
-        (push frame frame-list))
-      (when (eq (elt frame 1)
-                'buttercup--debugger)
-        (setq in-program-stack t))
-      ;; keep frames until one of the known functions are found, after
-      ;; this is just the buttercup framework and not interesting for
-      ;; users incorrect for testing buttercup. Some frames before the
-      ;; function also have to be discarded
-      (cl-labels ((tree-find (key tree)
-                             (cl-block tree-find
-                               (while (consp tree)
-                                 (let ((elem (pop tree)))
-                                   (when (or (and (consp elem)
-                                                  (tree-find key elem))
-                                             (eql key elem))
-                                     (cl-return-from tree-find t))))
-                               (cl-return-from tree-find
-                                 (and tree (eql tree key))))))
-        (when (and in-program-stack (tree-find 'buttercup--mark-stackframe frame))
-          (pop frame-list)
-          (cl-return frame-list)))))
+    ;; Keep frames until one if the end conditions is met. After
+    ;; this is just the buttercup framework and not interesting for
+    ;; users - except for testing buttercup.
+    (when (or
+           ;; When the error occurs in the calling of one of the
+           ;; wrapped expressions of an expect.
+           (buttercup--wrapper-fun-p (cadr frame))
+           ;; When an error happens in spec code but outside an expect
+           ;; statement
+           ;; buttercup--update-with-funcall
+           ;;  apply buttercup--funcall
+           ;;   buttercup--funcall   -   sets debugger
+           ;;    apply FUNCTION
+           ;;     FUNCTION  -- spec body function
+           ;;      condition-case   -- from buttercup-with-converted-ert-signals
+           ;;       (let ((buttercup--stackframe-marker 1))  -- the same
+           ;;        ACTUAL CODE
+           (and (null (car frame))
+                (eq 'let (cadr frame))
+                (equal '((buttercup--stackframe-marker 1)) (car (cddr frame)))
+                )
+           ;; TODO: What about :to-throw?
+           ;; buttercup--update-with-funcall (spec ...
+           ;;  apply buttercup--funcall
+           ;;   buttercup--funcall   -- sets the debugger
+           ;;    apply FUNCTION
+           ;;     FUNCTION -- spec body function
+           ;;      condition-case -- from buttercup-with-converted-ert-signals
+           ;;       (let ((buttercup--stackframe-marker 1))
+           ;;        (buttercup-expect
+           ;;         (buttercup--apply-matcher
+           ;;          (apply to-throw-matcher
+           ;;           (to-throw-matcher
+           ;;             We need a new debugger here, the
+           ;;             condition-case can not be used to collect
+           ;;             backtrace.
+           ;; When the error happens in the matcher function
+           ;;  (buttercup-expect
+           ;;   (buttercup--apply-matcher
+           ;;    (apply some-kind-of-function
+           ;;     (matcher
+           ;;      ACTUAL CODE
+           (and (eq 'buttercup--apply-matcher (cadr frame))
+                ;; The two preceeding frames are not of user interest
+                (pop frame-list) (pop frame-list)
+                ;; Add a fake frame for the matcher function
+                (push (cons t
+                            (cons (car (cddr frame))
+                                  (mapcar (lambda (x)
+                                            (if (buttercup--wrapper-fun-p x)
+                                                (buttercup--enclosed-expr x)
+                                              x))
+                                          (cadr (cddr frame)))))
+                      frame-list))
+           ;; TODO: What about signals in before and after blocks?
+           ;; BEFORE-EACH:
+           ;; buttercup--run-suite
+           ;;  (let* ...
+           ;;   (dolist (f (buttercup-suite-before-all ...
+           ;;    (buttercup--update-with-funcall suite f
+           ;;     (apply buttercup--funcall
+           ;;      (buttercup-funcall f
+           ;;       (f)
+           ;; Currently, buttercup silently ignores error in
+           ;; (before|after)-(all|each). As long as that is the case,
+           ;; there is nothing we can do about stacktraces.
+           )
+      (cl-return frame-list))
+    (push frame frame-list)))
 
 (defun buttercup--format-stack-frame (frame &optional style)
   "Format stack FRAME according to STYLE.
@@ -2193,7 +2228,9 @@ Specifically, `ert-test-failed' is converted to
 `buttercup-pending'."
   (declare (indent 0))
   `(condition-case err
-       (progn ,@body)
+       (let ((buttercup--stackframe-marker 1))
+         (ignore buttercup--stackframe-marker)
+         ,@body)
      (ert-test-failed
       (buttercup-fail "%S" err))
      (ert-test-skipped
